==> knips_0.6.20091202-1.tar.gz: knips-0.6.20091202/src/Control.cpp <==
/*
  Knips - A simple Camera program.
  Copyright (C) 2007  Tim Teulings

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Control.h"

#include <cstring>
#include <fstream>
#include <iostream>

// For video4linux access
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/videodev.h>
#include <linux/videodev2.h>

#include <gst/interfaces/colorbalance.h>
#include <gst/interfaces/xoverlay.h>
#include <gst/interfaces/videoorientation.h>
#include <gst/video/video.h>

#include <Lum/Base/String.h>
#include <Lum/Base/Util.h>

#include <Lum/Dialog.h>

#include <Lum/OS/Display.h>
#include <Lum/OS/EventLoopGMainLoop.h>

#include <Lum/OS/X11/Display.h>

#define CAMERA_ORIENTATION_FILE "/sys/devices/platform/gpio-switch/cam_turn/state"

class Device
{
public:
  std::string name;
  std::string device;
  std::string gstSink;
};

static gboolean cb_have_data(GstElement* /*element*/,
                             GstBuffer *buffer,
                             GstPad* /*pad*/,
                             gpointer user_data)
{
  static_cast<VideoControl*>(user_data)->GotImage(static_cast<unsigned char*>(GST_BUFFER_DATA(buffer)));

  return true;
}

gboolean GstCallback(GstBus* bus, GstMessage* message, gpointer user_data)
{
  static_cast<VideoControl*>(user_data)->OnGstMessage(message);

  return true;
}

void GetV4LDevices(std::list<Device>& devices)
{
  std::string devicePrefix="/dev/video";
  size_t      deviceIndex=0;

  devices.clear();

  while (true) {
    std::string             deviceName=devicePrefix+Lum::Base::NumberToString(deviceIndex);
    int                     handle;
    struct video_capability caps1;
    struct v4l2_capability  caps2;

    handle=open(deviceName.c_str(),O_RDWR);

    if (handle==-1) {
      return;
    }

    std::cout << "Detected device " << deviceName << std::endl;

    std::cout << "  Checking V4L2 compatibility..." << std::endl;

    if (ioctl(handle,VIDIOC_QUERYCAP,&caps2)!=-1) {
      std::cout << "    Got capabilities...\"" << caps2.driver << "\" \"" << caps2.card << "\" \"" << caps2.bus_info << "\"" << std::endl;

      if (caps2.capabilities & V4L2_CAP_VIDEO_CAPTURE) {
        std::cout << "    Video source can be captured..." << std::endl;

        size_t index=0;

        while (true) {
          struct v4l2_input input;

          input.index=index;

          if (ioctl(handle,VIDIOC_ENUMINPUT,&input)!=-1) {
            if (input.type & V4L2_INPUT_TYPE_CAMERA) {
              std::cout << "    Input " << index << " is of type camera!" << std::endl;

              Device device;

              device.name=(const char*)caps2.card;
              device.device=deviceName;
              device.gstSink="v4l2src";

              devices.push_back(device);
            }

            index++;
          }
          else {
            break;
          }
        }
      }
    }

    std::cout << "  Checking V4L compatibility..." << std::endl;

    if (ioctl(handle,VIDIOCGCAP,&caps1)!=-1) {
      std::cout << "    Got capabilities...\"" << caps1.name << "\"" << std::endl;

      if (caps1.type & VID_TYPE_CAPTURE) {
        std::cout << "    Video source can be captured..." << std::endl;
        for (size_t c=0; c<caps1.channels; c++) {
          struct video_channel channel;

          channel.channel=c;

          if (ioctl(handle,VIDIOCGCHAN,&channel)!=-1) {
            if (channel.type & VIDEO_TYPE_CAMERA) {
              std::cout << "    Channel " << c << " is of type camera!" << std::endl;

              Device device;

              device.name=caps1.name;
              device.device=deviceName;
              device.gstSink="v4lsrc";

              devices.push_back(device);
            }
          }
        }
      }
    }

    close(handle);

    std::cout << "done." << std::endl;

    deviceIndex++;
  }
}

VideoControl::VideoControl()
: pipeline(NULL),videoSink(NULL),imageSink(NULL),
  source(NULL),
  videoWidth(0),videoHeight(0),
  storeImage(false),done(false),
  gstMsg(new Lum::Model::Action()),
  knipsed(new Lum::Model::Action()),
  timer(new Lum::Model::Action())
{
  SetDestroyOnCleanup(false);

  Observe(gstMsg);
  Observe(timer);
}

VideoControl::~VideoControl()
{
  if (pipeline!=NULL) {
    gst_element_set_state(pipeline,GST_STATE_NULL);

    while (true) {
      GstState current;

      gst_element_get_state(pipeline,&current,NULL,GST_CLOCK_TIME_NONE);
      if (current==GST_STATE_NULL) {
        break;
      }
      Lum::Base::SystemTime::Sleep(1);
    }

    gst_object_unref(pipeline);
    pipeline=NULL;
  }

  if (source!=NULL) {
    g_source_destroy(source);
    g_source_unref(source);
  }
}

void VideoControl::CheckCameraOrientation()
{
  std::ifstream file;
  std::string   value;

  file.open(CAMERA_ORIENTATION_FILE,std::ios::in);

  if (!file) {
    return;
  }

  std::getline(file,value);

  file.close();

  if (gst_element_implements_interface(videoSource,GST_TYPE_VIDEO_ORIENTATION)) {
    GstVideoOrientation *orientation=GST_VIDEO_ORIENTATION(videoSource);

    gst_video_orientation_set_vflip(orientation,value=="active");
    gst_video_orientation_set_hflip(orientation,value=="active");
  }
}

void VideoControl::GotImage(unsigned char* data)
{
  Lum::OS::MutexGuard guard(mutex);

  if (!storeImage || videoWidth==0 || videoHeight==0) {
    return;
  }

  Lum::Images::Pixel *pixel;

  image=Lum::Images::Factory::factory->CreateImage();

  pixel=new Lum::Images::Pixel[videoWidth*videoHeight];

  for (size_t i=0; i<videoWidth*videoHeight; i++) {
    pixel[i].r=data[i*3];
    pixel[i].g=data[i*3+1];
    pixel[i].b=data[i*3+2];
    pixel[i].a=255;
  }

  image->SetData(videoWidth,videoHeight,false,pixel);

  storeImage=false;

  Lum::OS::display->QueueActionForAsyncNotification(knipsed.Get());
}

void VideoControl::SetKnipsAction(Lum::Model::Action* action)
{
  knips=action;
}

Lum::Model::Action* VideoControl::GetKnipsedAction() const
{
  return knipsed.Get();
}

void VideoControl::GetImage(Lum::Images::ImageRef &image)
{
  Lum::OS::MutexGuard guard(mutex);

  image=this->image;
}

void VideoControl::CalcSize()
{
  width=160;
  height=120;

  Observe(dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())->GetOpenedAction());

  minWidth=width;
  minHeight=height;

  Lum::OS::X11::SubWindow::CalcSize();
}

void VideoControl::Layout()
{
  if (!done) {
    GError            *error=NULL;
    std::list<Device> devices;
    std::string       videoSinkName;
    std::string       launchString;

    gst_init(NULL,NULL);

    GetV4LDevices(devices);

    if (devices.size()>0) {
      videoSinkName=devices.front().gstSink+" device="+devices.front().device;
    }
    else {
      videoSinkName="videotestsrc";
    }

    launchString=videoSinkName+" "+
                 "name=videosource ! "
                 "video/x-raw-rgb,width=640,height=480 ! "
                 "ffmpegcolorspace ! "
                 "tee name=tee tee. ! "
                 "queue ! "
                 "{ xvimagesink name=videosink } "
                 "tee. ! "
                 "queue ! "
                 "{ ffmpegcolorspace ! "
                 "video/x-raw-rgb,bpp=24,depth=24,frameRate=(fraction)15/1 ! "
                 "fakesink name=imagesink }";

//                 "video/x-raw-rgb,bpp=24,depth=24,framerate=(fraction)8/1 ! "

    std::cout << "Launch string is: \""+launchString+"\"" << std::endl;

    // Pipeline elements

    pipeline=::gst_parse_launch(launchString.c_str(),
                                &error);

    if (pipeline==NULL) {
      std::cerr << "Cannot open video source: " << error << std::endl;
      done=true;
      return;
    }

    videoSource=gst_bin_get_by_name(GST_BIN(pipeline),"videosource");
    videoSink=gst_bin_get_by_name(GST_BIN(pipeline),"videosink");
    imageSink=gst_bin_get_by_name(GST_BIN(pipeline),"imagesink");

    assert(videoSource!=NULL);
    assert(videoSink!=NULL);

    if (imageSink!=NULL) {
      assert(imageSink!=NULL);

      g_object_set(G_OBJECT(imageSink),"signal-handoffs",TRUE,NULL);
      g_signal_connect(imageSink,"handoff",G_CALLBACK(cb_have_data),this);
    }

    Lum::OS::GMainLoopEventLoop* eventLoop=dynamic_cast<Lum::OS::GMainLoopEventLoop*>(Lum::OS::display->GetEventLoop());
    GstBus*                      bus=gst_element_get_bus(pipeline);

    source=gst_bus_create_watch(bus);
    g_source_set_callback(source,(GSourceFunc)GstCallback,this,NULL);
    g_source_attach(source,eventLoop->context);
    gst_object_unref(GST_OBJECT(bus));

    gst_element_set_state(pipeline,GST_STATE_PLAYING);

    GstPad *videoSourcePad;

    videoSourcePad=gst_element_get_static_pad(videoSource,"src");

    if (videoSourcePad!=NULL) {
      gint w,h;

      if (gst_video_get_size(videoSourcePad,&w,&h)) {
        videoWidth=w;
        videoHeight=h;
        std::cout << "Image dimension: " << videoWidth << "x" << videoHeight << std::endl;
      }

      gst_object_unref(videoSourcePad);
    }

    Lum::OS::display->AddTimer(1,0,timer);

    done=true;
  }

  Lum::OS::X11::SubWindow::Layout();
}

void VideoControl::RequestImage()
{
  Lum::OS::MutexGuard guard(mutex);

  image=NULL;
  storeImage=true;
}

void VideoControl::CancelRequestImage()
{
  Lum::OS::MutexGuard guard(mutex);

  storeImage=false;
  image=NULL;
}

void VideoControl::OnGstMessage(GstMessage* message)
{
  //std::cout << "Message: " << gst_message_type_get_name(GST_MESSAGE_TYPE(message)) << std::endl;

  if (GST_MESSAGE_TYPE(message)==GST_MESSAGE_ERROR) {
    GError *error=NULL;
    char   *text=NULL;

    gst_message_parse_error(message,&error,&text);

    std::cerr << "gstreamer ERROR: " << text << std::endl;

    g_error_free(error);
    g_free(text);
  }
  else if (GST_MESSAGE_TYPE(message)==GST_MESSAGE_WARNING) {
    GError *error=NULL;
    char   *text=NULL;

    gst_message_parse_warning(message,&error,&text);

    std::cerr << "gstreamer WARNING: " << text << std::endl;

    g_error_free(error);
    g_free(text);
  }
  else if (message->src==GST_OBJECT(pipeline) && GST_MESSAGE_TYPE(message)==GST_MESSAGE_STATE_CHANGED) {
    GstState newState;

    gst_message_parse_state_changed(message,NULL,&newState,NULL);

    if (newState==GST_STATE_PLAYING || newState==GST_STATE_PAUSED) {
      knips->Enable();
    }
    else {
      knips->Disable();
    }
  }
}

bool VideoControl::HandleMouseEvent(const Lum::OS::MouseEvent& event)
{
  if (!knips->IsEnabled()) {
    return false;
  }

  if (event.type==Lum::OS::MouseEvent::down &&
      PointIsIn(event) &&
      event.button==Lum::OS::MouseEvent::button1) {
      if (!knips->IsStarted()) {
        knips->Start();
        return true;
      }
    }
    else if (event.IsGrabEnd()) {
      if (knips->IsStarted()) {
        if (PointIsIn(event)) {
            knips->Finish();
        }
        else {
          knips->Cancel();
        }
      }
    }

  return false;
}

void VideoControl::Resync(Lum::Base::Model* model, const Lum::Base::ResyncMsg& msg)
{
  if (model==timer && timer->IsFinished()) {
    CheckCameraOrientation();
    Lum::OS::display->AddTimer(1,0,timer);
  }
  else if (GetWindow()!=NULL &&
           dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())!=NULL &&
           model==dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())->GetOpenedAction() &&
           dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())->GetOpenedAction()->IsFinished() &&
           videoSink!=NULL) {
    gst_x_overlay_set_xwindow_id(GST_X_OVERLAY(videoSink),window);
  }

  Object::Resync(model,msg);
}

==> knips_0.6.20100104-2.tar.gz: knips-0.6.20100104/src/Control.cpp <==
/*
  Knips - A simple Camera program.
  Copyright (C) 2007  Tim Teulings

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Control.h"

#include <cassert>
#include <cstring>
#include <fstream>
#include <iostream>

// For video4linux access
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/videodev.h>
#include <linux/videodev2.h>

#include <gst/interfaces/colorbalance.h>
#include <gst/interfaces/xoverlay.h>
#include <gst/interfaces/videoorientation.h>
#include <gst/video/video.h>

#include <Lum/Base/String.h>
#include <Lum/Base/Util.h>

#include <Lum/Dialog.h>

#include <Lum/OS/Display.h>
#include <Lum/OS/EventLoopGMainLoop.h>

#include <Lum/OS/X11/Display.h>

#include "config.h"

#define CAMERA_ORIENTATION_FILE "/sys/devices/platform/gpio-switch/cam_turn/state"

class Device
{
public:
  std::string name;
  std::string device;
  std::string gstSink;
};

static gboolean cb_have_data(GstElement* element,
                             GstBuffer *buffer,
                             GstPad* /*pad*/,
                             gpointer user_data)
{
  VideoControl *control=static_cast<VideoControl*>(user_data);

  assert(control!=NULL);

  control->GotImage(element,
                    static_cast<unsigned char*>(GST_BUFFER_DATA(buffer)));

  return true;
}

gboolean GstCallback(GstBus* bus, GstMessage* message, gpointer user_data)
{
  VideoControl *control=static_cast<VideoControl*>(user_data);

  assert(control!=NULL);

  control->OnGstMessage(message);

  return true;
}

void GetV4LDevices(std::list<Device>& devices)
{
  std::string devicePrefix="/dev/video";
  size_t      deviceIndex=0;

  devices.clear();

  while (true) {
    std::string             deviceName=devicePrefix+Lum::Base::NumberToString(deviceIndex);
    int                     handle;
    struct video_capability caps1;
    struct v4l2_capability  caps2;

    handle=open(deviceName.c_str(),O_RDWR);

    if (handle==-1) {
      return;
    }

    std::cout << "Detected device " << deviceName << std::endl;

    std::cout << "  Checking V4L2 compatibility..." << std::endl;

    if (ioctl(handle,VIDIOC_QUERYCAP,&caps2)!=-1) {
      std::cout << "    Got capabilities...\"" << caps2.driver << "\" \"" << caps2.card << "\" \"" << caps2.bus_info << "\"" << std::endl;

      if (caps2.capabilities & V4L2_CAP_VIDEO_CAPTURE) {
        std::cout << "    Video source can be captured..." << std::endl;

        size_t index=0;

        while (true) {
          struct v4l2_input input;

          input.index=index;

          if (ioctl(handle,VIDIOC_ENUMINPUT,&input)!=-1) {
            if (input.type & V4L2_INPUT_TYPE_CAMERA) {
              std::cout << "    Input " << index << " is of type camera!" << std::endl;

              Device device;

              device.name=(const char*)caps2.card;
              device.device=deviceName;
              device.gstSink="v4l2src";

              devices.push_back(device);
            }

            index++;
          }
          else {
            break;
          }
        }
      }
    }

    std::cout << "  Checking V4L compatibility..." << std::endl;

    if (ioctl(handle,VIDIOCGCAP,&caps1)!=-1) {
      std::cout << "    Got capabilities...\"" << caps1.name << "\"" << std::endl;

      if (caps1.type & VID_TYPE_CAPTURE) {
        std::cout << "    Video source can be captured..." << std::endl;
        for (size_t c=0; c<caps1.channels; c++) {
          struct video_channel channel;

          channel.channel=c;

          if (ioctl(handle,VIDIOCGCHAN,&channel)!=-1) {
            if (channel.type & VIDEO_TYPE_CAMERA) {
              std::cout << "    Channel " << c << " is of type camera!" << std::endl;

              Device device;

              device.name=caps1.name;
              device.device=deviceName;
              device.gstSink="v4lsrc";

              devices.push_back(device);
            }
          }
        }
      }
    }

    close(handle);

    std::cout << "done." << std::endl;

    deviceIndex++;
  }
}

VideoControl::VideoControl()
: pipeline(NULL),videoSink(NULL),imageSink(NULL),
  videoWidth(0),videoHeight(0),
  storeImage(false),done(false),
  gstMsg(new Lum::Model::Action()),
  knipsed(new Lum::Model::Action()),
  timer(new Lum::Model::Action())
{
  SetDestroyOnCleanup(false);

  Observe(gstMsg);
  Observe(timer);
}

VideoControl::~VideoControl()
{
  if (pipeline!=NULL) {
    std::cout << "Stoping pipeline..." << std::endl;

    gst_element_set_state(pipeline,GST_STATE_NULL);

    while (true) {
      GstState current;

      gst_element_get_state(pipeline,&current,NULL,GST_CLOCK_TIME_NONE);
      if (current==GST_STATE_NULL) {
        break;
      }
      Lum::Base::SystemTime::Sleep(1);
    }

    std::cout << "Pipeline stoped." << std::endl;

    gst_object_unref(pipeline);
    pipeline=NULL;
  }
}

void VideoControl::CheckCameraOrientation()
{
  std::ifstream file;
  std::string   value;

  file.open(CAMERA_ORIENTATION_FILE,std::ios::in);

  if (!file) {
    return;
  }

  std::getline(file,value);

  file.close();

  if (gst_element_implements_interface(videoSource,GST_TYPE_VIDEO_ORIENTATION)) {
    GstVideoOrientation *orientation=GST_VIDEO_ORIENTATION(videoSource);

    gst_video_orientation_set_vflip(orientation,value=="active");
    gst_video_orientation_set_hflip(orientation,value=="active");
  }
}

void VideoControl::GotImage(GstElement* imageSink,
                            unsigned char* data)
{
  Lum::OS::MutexGuard guard(mutex);

  g_signal_handler_disconnect(G_OBJECT(imageSink),
                              callbackId);
  callbackId=0;

  if (!storeImage || videoWidth==0 || videoHeight==0) {
    storeImage=false;
    return;
  }

  Lum::Images::Pixel *pixel;

  image=Lum::Images::Factory::factory->CreateImage();

  pixel=new Lum::Images::Pixel[videoWidth*videoHeight];

  for (size_t i=0; i<videoWidth*videoHeight; i++) {
    pixel[i].r=data[i*3];
    pixel[i].g=data[i*3+1];
    pixel[i].b=data[i*3+2];
    pixel[i].a=255;
  }

  image->SetData(videoWidth,videoHeight,false,pixel);

  Lum::OS::display->QueueActionForAsyncNotification(knipsed.Get());
}

void VideoControl::SetKnipsAction(Lum::Model::Action* action)
{
  knips=action;
}

Lum::Model::Action* VideoControl::GetKnipsedAction() const
{
  return knipsed.Get();
}

void VideoControl::GetImage(Lum::Images::ImageRef &image)
{
  Lum::OS::MutexGuard guard(mutex);

  image=this->image;
}

void VideoControl::CalcSize()
{
  width=160;
  height=120;

  Observe(dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())->GetOpenedAction());

  minWidth=width;
  minHeight=height;

  Lum::OS::X11::SubWindow::CalcSize();
}

void VideoControl::Layout()
{
  if (!done) {
    GError            *error=NULL;
    std::list<Device> devices;
    std::string       videoSinkName;
    std::string       launchString;

    done=true;

    gst_init(NULL,NULL);

#if defined(HAVE_LIB_HILDON)
    videoSinkName="v4l2camsrc";
#else
    GetV4LDevices(devices);

    if (devices.size()>0) {
      videoSinkName=devices.front().gstSink+" device="+devices.front().device;
    }
    else {
      videoSinkName="videotestsrc";
    }
#endif

    launchString=videoSinkName+" "+
                 "name=videosource ! "
                 "ffmpegcolorspace ! "
                 "video/x-raw-yuv,width=640,height=480,bpp=24,framerate=(fraction)25/1 ! "
                 "tee name=t ! queue ! "
                 "{ xvimagesink name=videosink } "
                 "t. ! queue ! "
                 "{ ffmpegcolorspace ! "
                 "video/x-raw-rgb,width=640,height=480,bpp=24 ! "
                 "fakesink name=imagesink }";
/*
    launchString=videoSinkName+" "+
                 "name=videosource num-buffers=1 ! "
                 "ffmpegcolorspace ! "
                 "video/x-raw-yuv,width=640,height=480,bpp=24,framerate=(fraction)25/1 ! "
                 "xvimagesink name=videosink";*/

    std::cout << "Launch string is: \""+launchString+"\"" << std::endl;

    // Pipeline elements
    pipeline=::gst_parse_launch(launchString.c_str(),
                                &error);

    if (pipeline==NULL) {
      std::cerr << "Cannot open video source: " << error << std::endl;
      done=true;
      return;
    }

    GstBus *bus;

    bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
    gst_bus_add_watch(bus,(GstBusFunc)GstCallback,this);
    gst_object_unref(GST_OBJECT(bus));

    videoSource=gst_bin_get_by_name(GST_BIN(pipeline),"videosource");
    videoSink=gst_bin_get_by_name(GST_BIN(pipeline),"videosink");
    imageSink=gst_bin_get_by_name(GST_BIN(pipeline),"imagesink");

    assert(videoSource!=NULL);
    assert(videoSink!=NULL);

    if (imageSink!=NULL) {
      g_object_set(G_OBJECT(imageSink),"signal-handoffs",true,(char*)NULL);
    }

    gst_element_set_state(pipeline,GST_STATE_PLAYING);

    GstPad *videoSourcePad;

    videoSourcePad=gst_element_get_static_pad(videoSource,"src");

    if (videoSourcePad!=NULL) {
      gint w,h;

      if (gst_video_get_size(videoSourcePad,&w,&h)) {
        videoWidth=w;
        videoHeight=h;
        std::cout << "Image dimension: " << videoWidth << "x" << videoHeight << std::endl;
      }

      gst_object_unref(videoSourcePad);
    }

    Lum::OS::display->AddTimer(1,0,timer);
  }

  Lum::OS::X11::SubWindow::Layout();
}

void VideoControl::RequestImage()
{
  Lum::OS::MutexGuard guard(mutex);

  if (pipeline!=NULL) {
    image=NULL;
    storeImage=true;

    GstElement *imageSink;

    imageSink=gst_bin_get_by_name(GST_BIN(pipeline),
                                  "imagesink");

    if (imageSink!=NULL) {
      callbackId=g_signal_connect(G_OBJECT(imageSink),"handoff",
                                  G_CALLBACK(cb_have_data),this);
    }
  }
}

void VideoControl::CancelRequestImage()
{
  Lum::OS::MutexGuard guard(mutex);

  if (pipeline!=NULL) {
    GstElement *imageSink;

    imageSink=gst_bin_get_by_name(GST_BIN(pipeline),
                                  "imagesink");

    if (imageSink!=NULL) {
      g_signal_handler_disconnect(G_OBJECT(imageSink),
                                  callbackId);
      callbackId=0;
    }
  }

  storeImage=false;
  image=NULL;
}

void VideoControl::OnGstMessage(GstMessage* message)
{
  //std::cout << "Message: " << gst_message_type_get_name(GST_MESSAGE_TYPE(message)) << std::endl;

  if (GST_MESSAGE_TYPE(message)==GST_MESSAGE_ERROR) {
    GError *error=NULL;
    char   *text=NULL;

    gst_message_parse_error(message,&error,&text);

    std::cerr << "gstreamer ERROR: " << text << std::endl;

    g_error_free(error);
    g_free(text);
  }
  else if (GST_MESSAGE_TYPE(message)==GST_MESSAGE_WARNING) {
    GError *error=NULL;
    char   *text=NULL;

    gst_message_parse_warning(message,&error,&text);

    std::cerr << "gstreamer WARNING: " << text << std::endl;

    g_error_free(error);
    g_free(text);
  }
  else if (message->src==GST_OBJECT(pipeline) && GST_MESSAGE_TYPE(message)==GST_MESSAGE_STATE_CHANGED) {
    GstState newState;

    gst_message_parse_state_changed(message,NULL,&newState,NULL);

    if (newState==GST_STATE_PLAYING || newState==GST_STATE_PAUSED) {
      knips->Enable();
    }
    else {
      knips->Disable();
    }
  }
}

bool VideoControl::HandleMouseEvent(const Lum::OS::MouseEvent& event)
{
  if (!knips->IsEnabled()) {
    return false;
  }

  if (event.type==Lum::OS::MouseEvent::down &&
      PointIsIn(event) &&
      event.button==Lum::OS::MouseEvent::button1) {
      if (!knips->IsStarted()) {
        knips->Start();
        return true;
      }
    }
    else if (event.IsGrabEnd()) {
      if (knips->IsStarted()) {
        if (PointIsIn(event)) {
            knips->Finish();
        }
        else {
          knips->Cancel();
        }
      }
    }

  return false;
}

void VideoControl::Resync(Lum::Base::Model* model, const Lum::Base::ResyncMsg& msg)
{
  if (model==timer && timer->IsFinished()) {
    CheckCameraOrientation();
    Lum::OS::display->AddTimer(1,0,timer);
  }
  else if (GetWindow()!=NULL &&
           dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())!=NULL &&
           model==dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())->GetOpenedAction() &&
           dynamic_cast<Lum::Dialog*>(GetWindow()->GetMaster())->GetOpenedAction()->IsFinished()) {
    if (videoSink!=NULL) {
      std::cout << "Assigning window to overlay..." << std::endl;
      gst_x_overlay_set_xwindow_id(GST_X_OVERLAY(videoSink),window);
    }
    else {
      std::cerr << "Window opened but imageSink is not available!" << std::endl;
    }
  }

  Object::Resync(model,msg);
}
